#ifndef _VSIM_HITVECTOR_H_
#define _VSIM_HITVECTOR_H_
#include <vector>
#include <functional>
#include "types.h"

namespace vsim {
/**
 * The HitVector class implements an array interface towards a preallocated
 * buffer of known size, which may quickly be erased and reused.
 */
class HitVector {
public:
    /**
     * Construct a HitVector with an internal buffer of size maxlen.
     */
    HitVector(docid_t maxlen) :
        _buffer(maxlen),
        _len(0),
        _best(0)
    { }
    const Hit &operator[] (unsigned long a) const { return _buffer[a]; }
    /**
     * Add a new hit to the buffer.
     */
    inline void add(const Hit &h) { _buffer[_len++] = h; }

    /**
     * Sort the array according to the comparators defined in this class.
     */
    void sort(void);

    /**
     * Merge two vectors provided as arguments into this vector starting at the
     * current offset pointer.
     */
    void merge(const HitVector &, const HitVector &);
    
    /**
     * Used to compared values for radix sort.
     */
    class ValueCompare : public std::binary_function<Hit, Hit, bool> {
    public:
        bool operator() (const Hit &lhs, const Hit &rhs) const {
            return (lhs.second > rhs.second);
        }
    };

    /**
     * Used to retrieve sort value
     */
    class ValueRadix {
    public:
        uint64_t operator() (const Hit & h) const {
            return (uint64_t)(~h.second);
        }
    };


    /**
     * Various methods defined to be able to be used as an STL container.
     */
    inline Hit &back(void) { return (_buffer[_len - 1]); }
    inline std::vector<Hit>::iterator begin(void) { return (_buffer.begin()); }
    inline std::vector<Hit>::iterator end(void) { return (_buffer.begin() + _len); }
    inline unsigned long size(void) const { return (_len); }
    inline bool empty(void) { return (_len == 0); }
    inline void reset(void) { _len = 0; }
    void printHits(void) const;
private:
    std::vector<Hit> _buffer;
    docid_t _maxlen;
    docid_t _len;
    docid_t _best;
};

} // end vsim
#endif
