#include <iostream>
#include <cstring>

#include "vsim.h"
#include "vsimconfig.h"
#include "vsimreport.h"
#include "types.h"


void
usage(const char *prog)
{
    std::cerr << "Usage: " << prog << " [OPTION]... simlength" << std::endl;
    std::cerr << "OPTION may be one of these:" << std::endl;
    
    struct option {
        const char *name;
        const char *description;
    };
    struct option options[] = {
        {"a", "The number of attributes per document"},
        {"c", "Maximum number of queryhandlers"},
        {"d", "Maximum document id"},
        {"h", "Maximum number of hits returned per query {heap size)"},
        {"i", "The probability distribution to apply to the slots (binomial, poisson, geometric or gauss)"},
        {"l", "The number of slots to initially allocate per partition"},
        {"n", "A static number of hits used per query (unless using QRS log files)"},
        {"o", "Sample file to store simulation data"},
        {"p", "Number of partitions"},
        {"q", "Prefix of QRS log files used to generate hit counts"},
        {"r", "A comma-separated list of attribute vector indices (or a range of indices) used for rank calculation"},
        {"s", "The number of rounds to run the simulation"},
        {"t", "Number of clients to spawn"},
        {NULL, NULL}
    };

    unsigned int i = 0;
    while (options[i].name != NULL && options[i].description != NULL) {
        std::cerr << "\t-" << options[i].name << "\t\t" << options[i].description << std::endl;
        i++;
    }
}


int
main(int argc, char **argv)
{
    int opt;
    bool set_attr_range = false;
    vsim::docid_t numpart, numslots, maxdocid, hitheapsize, numhits, maxdocid_tmp;
    unsigned long numattr, maxconn, clients, numrounds;
    std::string attrrange, qrlog, samplefile, distribution;
    vsim::AttributeIndexVector aiv;

    // Default values
    numpart = numslots = maxconn = 1;
    clients = 1;
    maxdocid = hitheapsize = numhits = maxdocid_tmp = 1000;
    numattr = numrounds = 1;
    aiv.push_back(0);

    while ((opt = getopt(argc, argv, "a:c:d:h:i:l:m:n:o:p:q:r:s:t:")) != -1) {
        switch (opt) {
        case 'a':
            numattr = strtol(optarg, NULL, 10);
            break;
        case 'c':
            maxconn = strtol(optarg, NULL, 10);
            break;
        case 'd':
            maxdocid_tmp = strtol(optarg, NULL, 10);
            break;
        case 'h':
            hitheapsize = strtol(optarg, NULL, 10);
            break;
        case 'i':
            distribution = std::string(optarg);
            break;
        case 'l':
            numslots = strtol(optarg, NULL, 10);
            break;
        case 'n':
            numhits = strtol(optarg, NULL, 10);
            break;
        case 'o':
            samplefile = std::string(optarg);
            break;
        case 'p':
            numpart = strtol(optarg, NULL, 10);
            break;
        case 'q':
            qrlog = std::string(optarg);
            break;
        case 'r':
            attrrange = std::string(optarg);
            set_attr_range = true;
            break;
        case 's':
            numrounds = strtol(optarg, NULL, 10);
            break;
        case 't':
            clients = strtol(optarg, NULL, 10);
            break;
        default:
            usage(argv[0]);
            exit(-1);
            break;
        }
    }
    if (optind >= argc) {
        std::cerr << "Expected argument after options" << std::endl;
        usage(argv[0]);
        exit(-1);
    }
    
    unsigned long simtime = 0;
    simtime = strtol(argv[optind], NULL, 10);

    if (set_attr_range) {
        aiv.clear();
        // Trim whitespace
        for (std::string::iterator it = attrrange.begin();
             it != attrrange.end();
             it++) {

            if ((*it) == ' ' || (*it) == '\t')
                attrrange.erase(it);
        }
        size_t first = 0;
        size_t last = std::string::npos;
        do {
            last = attrrange.find(',', first);
            std::string attr(attrrange.substr(first, (last - first)));
            // Find the '-' delimiter;
            size_t delimpos = attr.find('-');
            if (delimpos != std::string::npos) {
                std::string tmp(attr.substr(0, delimpos));
                unsigned long startattr = strtol(tmp.c_str(), NULL, 10);
                tmp = attr.substr(delimpos + 1, (attr.length() - delimpos));
                unsigned long lastattr = strtol(tmp.c_str(), NULL, 10);
                if (startattr >= 0 && lastattr <= numattr) {
                    for (unsigned int i = startattr; i < lastattr; i++) {
                        aiv.push_back(i);
                    }
                }
            } else {
                unsigned long attrnum = strtol(attr.c_str(), NULL, 10);
                if (attrnum >= 0 && attrnum < numattr)
                    aiv.push_back(attrnum);
            }

            first = last + 1;
        } while (last != std::string::npos);
    }
    maxdocid = ROUND(maxdocid_tmp, numpart);

    vsim::VSimConfig cfg(numpart, numslots, maxconn, clients, maxdocid, numhits,
        hitheapsize, numattr, aiv, vsim::slotDistTextToEnum(distribution),
        simtime, numrounds, qrlog);
    vsim::VSim sim(cfg);
    sim.run(samplefile);
    return 0;
}

