#include <iostream>
#include <algorithm>
#include "atomic.h"
#include "query.h"
#include "types.h"
#include "misc.h"
#include "vsimtime.h"
#include "hitcollector.h"

namespace vsim {

Query::Query(const PartitionSpecSP &ps, docid_t maxdocid, partid_t numpart, docid_t heapsize, docid_t numhits) :
    _spec(ps),
    _maxdocid(maxdocid),
    _numpart(numpart),
    _numhits(numhits),
    _result(heapsize, numpart),
    _hs(numpart, 0),
    _finished(2),
    _insertPartitionQueue(numpart),
    _removePartitionQueue(numpart),
    _stopFetcher(numpart)
{
}

docid_t
Query::getHits(const Slot &slot)
{
    return (docid_t)((double)_numhits * slot.prob);
}

const PartitionRange &
Query::getRange(partid_t partid)
{
    return (*_spec)[partid];
}

void
Query::notify(partid_t partid, unsigned long numhits)
{
    _hs[partid] = numhits;
    stopFetcher(partid);
    _done.enqueue(partid);
}

void
Query::merge(void)
{
    for (partid_t nummerged = 0; nummerged < _numpart; nummerged++) {
        partid_t part = _done.dequeue();
        _result.mergeHits(part);
    }
    _finished.wait();
}

void
Query::wait(void)
{
    _finished.wait();
}

void
Query::reportStats(VSimStat &st)
{
    std::vector<uint64_t> fetchertimes(_numpart);
    std::vector<uint64_t> partqueuetimes(_numpart);
    uint64_t maxfetchertime = 0;
    for (partid_t i = 0; i < _numpart; i++) {
        fetchertimes[i] = _removePartitionQueue[i].getDiffMicroSeconds(_stopFetcher[i]);
        if (fetchertimes[i] > maxfetchertime)
            maxfetchertime = fetchertimes[i];
        partqueuetimes[i] = _insertPartitionQueue[i].getDiffMicroSeconds(_removePartitionQueue[i]);
    }
    unsigned long handlertime = _removeQueryQueue.getDiffMicroSeconds(_stopHandler);
    unsigned long queryqueuetime = _insertQueryQueue.getDiffMicroSeconds(_removeQueryQueue);
    st.report(fetchertimes, partqueuetimes, maxfetchertime, queryqueuetime, handlertime);
    st.reportHits(_hs);
}

} // end vsim
