#ifndef _VSIM_QUERY_H_
#define _VSIM_QUERY_H_
#include <boost/shared_ptr.hpp>
#include "types.h"
#include "resultbuffer.h"
#include "misc.h"
#include "vsimstat.h"
#include "vsimtime.h"
#include "barrier.h"
#include "mergequeue.h"

namespace vsim {

class QueryHandler;

/**
 * A query contains the information needed to generate document id hits and
 * buffers to store the hits.
 */
class Query
{
public:
    typedef boost::shared_ptr<Query> SP;

    Query(const PartitionSpecSP &, docid_t, partid_t, docid_t, docid_t);
    /**
     * Get the number of hits calculated from the slots probability and the
     * number of hits the query should get.
     */
    docid_t getHits(const Slot &);

    /**
     * Get the specific partition range for the partition given as argument.
     */
    const PartitionRange &getRange(partid_t);

    inline HitCollector &getCollector(partid_t part) { return _result.getHandle(part); }

    /**
     * Notify that a partition is finished collecting hits and ready for
     * merging.
     */
    void notify(partid_t, unsigned long);

    /**
     * Perform the merge of the results from each partition. Use the merge queue
     * to serialize merging.
     */
    void merge(void);

    /**
     * For the client, to wait until a query is done.
     */
    void wait(void);

    /**
     * Report all statistics collected for this query to the VSimStat object.
     */
    void reportStats(VSimStat &);

    /**
     * Methods used to record time stamps for query events.
     */
    void insertQueryQueue(void) { _insertQueryQueue.recordTime(); }
    void insertPartitionQueue(partid_t partid) { _insertPartitionQueue[partid].recordTime(); }
    void removePartitionQueue(partid_t partid) { _removePartitionQueue[partid].recordTime(); }
    void stopFetcher(partid_t partid) { _stopFetcher[partid].recordTime(); }
    void removeQueryQueue(void) { _removeQueryQueue.recordTime(); }
    void stopHandler(void) { _stopHandler.recordTime(); }

private:
    PartitionSpecSP _spec;
    docid_t _maxdocid;
    partid_t _numpart;
    docid_t _numhits;
    ResultBuffer _result;
    HitStat _hs;
    MergeQueue _done;
    Barrier _finished;

    // Statistics kept for us
    VSimTime _insertQueryQueue;
    std::vector<VSimTime> _insertPartitionQueue;
    std::vector<VSimTime> _removePartitionQueue;
    std::vector<VSimTime> _stopFetcher;
    VSimTime _removeQueryQueue;
    VSimTime _stopHandler;
};

} // end vsim

#endif
