#include <iostream>
#include <fstream>
#include <cstdlib>
#include "query.h"
#include "querygenerator.h"

namespace vsim {

QueryGenerator::QueryGenerator(QueryQueue &queue, const PartitionSpecSP &spec,
    docid_t maxdocid, unsigned long heapsize) :
    _queue(queue),
    _done(2),
    _shutdown(false),
    _spec(spec),
    _maxdocid(maxdocid),
    _heapsize(heapsize)
{}

QueryGenerator::~QueryGenerator()
{}

void
QueryGenerator::setPartitionSpec(const PartitionSpecSP &spec)
{
    _spec = spec;
}

void
QueryGenerator::run(void)
{
    _lock.lock();
    while (!_shutdown) {
        _lock.unlock();
        Query::SP qsp(genQuery());
        _queue.enqueue(qsp);
        qsp->wait();
        _lock.lock();
    }
    _lock.unlock();
    _done.wait();
}

void
QueryGenerator::shutdown(void)
{
    _lock.lock();
    _shutdown = true;
    _lock.unlock();
}

void
QueryGenerator::wait(void)
{
    _done.wait();
}

bool
QueryGenerator::getPreferredCPU(cpu_set_t &set)
{
    return false;
}

Query::SP
QueryGenerator::createQuery(docid_t numhits)
{
    return (Query::SP(new Query(_spec, _maxdocid, (*_spec).size(), _heapsize, numhits)));
}

StaticQueryGenerator::StaticQueryGenerator(QueryQueue &queue, const PartitionSpecSP &spec,
    docid_t maxdocid, unsigned long heapsize, docid_t numhits) :
    QueryGenerator(queue, spec, maxdocid, heapsize),
    _numhits(numhits)
{
}

Query::SP
StaticQueryGenerator::genQuery(void)
{
    return (createQuery(_numhits));
}

QRSLogQueryGenerator::QRSLogQueryGenerator(QueryQueue &queue,
    const PartitionSpecSP &spec, docid_t maxdocid, unsigned long heapsize,
    const std::string &qrlog) :

    QueryGenerator(queue, spec, maxdocid, heapsize)
{
    _input = fopen(qrlog.c_str(), "r");
}

QRSLogQueryGenerator::~QRSLogQueryGenerator()
{
    fclose(_input);
}

Query::SP
QRSLogQueryGenerator::genQuery(void)
{
    unsigned long numhits = 0;
    do {
        if (fgets(_buffer, buflen, _input) == NULL) {
            rewind(_input);
        }
        numhits = strtol(_buffer, NULL, 10);
    } while (numhits == 0);
    
#define MAXHITS 1000000
    if (numhits > MAXHITS)
        numhits = MAXHITS;
    return (createQuery(numhits));
}

} // end vsim
