#ifndef _VSIM_QUERYGENERATOR_H_
#define _VSIM_QUERYGENERATOR_H_
#include <boost/shared_ptr.hpp>
#include <boost/random/linear_congruential.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/random/exponential_distribution.hpp>
#include <boost/random/variate_generator.hpp>
#include <fstream>
#include <string>
#include "vsimtime.h"
#include "query.h"
#include "types.h"
#include "task.h"
#include "queryqueue.h"

namespace vsim {

/**
 * A Query Generator interface, capable of generating queries for use by the
 * Vespa simulator.
 */
class QueryGenerator : public Task
{
public:
    typedef boost::shared_ptr<QueryGenerator> SP;
    QueryGenerator(QueryQueue &, const PartitionSpecSP &, docid_t, unsigned long);
    virtual ~QueryGenerator();
    Query::SP createQuery(docid_t);
    void setPartitionSpec(const PartitionSpecSP &);
    void wait(void);

    void run(void);
    void shutdown(void);
    bool getPreferredCPU(cpu_set_t &);

    /**
     * Method that should be used by users of query generators, which must be
     * implemented by sub-classes.
     */
    virtual Query::SP genQuery(void) = 0;

private:
    QueryQueue &_queue;
    Mutex _lock;
    Barrier _done;
    bool _shutdown;


protected:
    PartitionSpecSP _spec;
    docid_t _maxdocid;
    unsigned long _heapsize;


};

/**
 * The static query generator generates a fixed number of hits for each
 * partition, and is mostly used to verify the models in the ideal cases.
 */
class StaticQueryGenerator : public QueryGenerator
{
public:
    StaticQueryGenerator(QueryQueue &, const PartitionSpecSP &, docid_t,
        unsigned long, docid_t);
    Query::SP genQuery(void);
private:
    unsigned long _numhits;
};

/**
 * The query log query generator is used when generating queries from a query
 * log file given as input.
 */
class QRSLogQueryGenerator : public QueryGenerator
{
public:
    QRSLogQueryGenerator(QueryQueue &, const PartitionSpecSP &, docid_t,
        unsigned long, const std::string &);
    ~QRSLogQueryGenerator();
    Query::SP genQuery(void);
private:
    FILE *_input;
    static const unsigned int buflen = 128;
    char _buffer[buflen];
};

} // end vsim

#endif
