#include <iostream>
#include <time.h>

#include "vsim.h"
#include "vsimstat.h"
#include "query.h"
#include "querygenerator.h"
#include "queryhandler.h"
#include "fetcher.h"
#include "threadpool.h"
#include "vsimstore.h"
#include "vsimconfig.h"
#include "vsimprofiler.h"
#include "vsimseeder.h"

namespace vsim {

void
VSim::run(const std::string &samplefile)
{
    VSimStore ds(_vcfg);
    VSimProfiler::stop();
    for (unsigned int i = 0; i < _vcfg.getNumberOfRounds(); i++) {
        seed(ds);
        ds.memlock();
        VSimProfiler::start();
        VSimReport report = doRun(ds);
        VSimProfiler::stop();
        ds.memunlock();
        report.printReport();
        // Output samples
        if (samplefile.size() > 0) {
            report.exportReport(samplefile.c_str());
        }
        sleep(1);
    }
}

void
VSim::seed(VSimStore &ds)
{
    std::vector<VSimSeeder::SP> seeders;
    PartitionSpecSP ps = _vcfg.createPartitionSpec();
    ThreadPool tp(ps->size() + 1);

    for (partid_t i = 0; i < (*ps).size(); i++) {
        PartitionRange &pr((*ps)[i]);
        docid_t start = pr[0].first;
        docid_t end = pr[pr.size() - 1].second;
        VSimSeeder::SP vssp(new VSimSeeder(ds, start, end, i));
        seeders.push_back(vssp);
        tp.runThread(vssp);
    }
    for (partid_t i = 0; i < seeders.size(); i++) {
        seeders[i]->shutdown();
    }
}

QueryGenerator::SP
VSim::createQueryGenerator(QueryQueue &queue, const PartitionSpecSP &spec,
    unsigned int id)
{
    std::string qrlog(_vcfg.getQRSLogFile());
    if (qrlog.size() > 0) {
        char buf[128];
        if (id < 10)
            snprintf(buf, 128, "0%u", id);
        else
            snprintf(buf, 128, "%u", id);
        qrlog += buf;
        std::cout << qrlog << std::endl;
        return (QueryGenerator::SP(new QRSLogQueryGenerator(queue, spec,
            _vcfg.getMaxDocId(), _vcfg.getHitHeapSize(), qrlog)));
    }
    return (QueryGenerator::SP(new StaticQueryGenerator(queue, spec,
        _vcfg.getMaxDocId(), _vcfg.getHitHeapSize(), _vcfg.getNumHits())));
}


VSimReport
VSim::doRun(VSimStore &ds)
{
    VSimStat st(_vcfg);
    PartitionSpecSP ps = _vcfg.createPartitionSpec();
    ThreadPool tp(300);

    std::cout << "Setting up simulation\n";
    std::vector<PartitionQueue> partitions;
    std::vector<Fetcher::SP> fetchers;
    std::vector<QueryGenerator::SP> clients;
    QueryQueue qq(_vcfg.getMaxConnections());
    std::vector<QueryHandler::SP> handlers;

    // Create partitions
    for (partid_t i = 0; i < _vcfg.getNumPartitions(); i++) {
        partitions.push_back(PartitionQueue(i));
    }

    // Create clients
    for (unsigned int i = 0; i < _vcfg.getNumClients(); i++) {
        QueryGenerator::SP qgsp(createQueryGenerator(qq, ps, i));
        clients.push_back(qgsp);
    }

    // The number of fetchers per partition depends on the number of partitions.
    unsigned int fetchers_per_part = _vcfg.getMaxConnections();
    for (partid_t i = 0; i < partitions.size(); i++) {
        for (unsigned int j = 0; j < fetchers_per_part; j++) {
            VSimTime vst;
            vst.recordTime();
            Fetcher::SP fsp(new Fetcher(ds, partitions[i], i,
                (vst.getTimeNanoSeconds() + j + i), j, _vcfg.getMaxDocId()));
            tp.runThread(fsp);
            fetchers.push_back(fsp);
        }
    }

    // Create query queue and queryhandlers
    for (unsigned int i = 0; i < _vcfg.getMaxConnections(); i++) {
        QueryHandler::SP qhsp(new QueryHandler(partitions, st, qq));
        tp.runThread(qhsp);
        handlers.push_back(qhsp);
    }

    std::cout << "Let threads settle down\n";
    sleep(2);
    std::cout << "Starting simulation\n";
    st.startSimulation();

#ifdef AUTOPARTITIONING
#define PARTCHECKNUM   100
    unsigned int partcheck = PARTCHECKNUM;
#endif
    // Start all clients
    for (unsigned int i = 0; i < clients.size(); i++) {
        tp.runThread(clients[i]);
    }
    st.sleepUntilFinished();
#ifdef AUTOPARTITIONING
            if (partitions.size() > 1) {
                partcheck--;
                if (partcheck <= 0) {
                    ps = st.updatePartitionSpec(ps);
                    partcheck = PARTCHECKNUM;
                }
            }
#endif
    std::cout << "Stopping simulation\n";
    for (unsigned int i = 0; i < clients.size(); i++) {
        clients[i]->shutdown();
    }
    for (unsigned int i = 0; i < clients.size(); i++) {
        clients[i]->wait();
    }
    qq.waitUntilEmpty();
    qq.waitUntilEmpty();
    qq.shutdown();
    for (unsigned int i = 0; i < handlers.size(); i++) {
        handlers[i]->shutdown();
    }
    for (partid_t i = 0; i < partitions.size(); i++) {
        partitions[i].waitUntilEmpty();
        partitions[i].shutdown();
    }
    for (partid_t i = 0; i < fetchers.size(); i++) {
        fetchers[i]->shutdown();
    }
    st.stopSimulation();
    std::cout << "Simulation finished\n";
    return st.generateReport(*ps);
}

} // end vsim
