#include <string>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cstring>
#include "vsimreport.h"
#include "simendian.h"
#include "misc.h"

namespace vsim {

#define WRITE64(fd, var) do {               \
    uint64_t tmp = vsimhtobe64((var));      \
    (fd).write((char *)&tmp, sizeof(tmp));  \
} while (0)

void
VSimReport::exportReport(const char *outfile)
{
    std::ofstream fd(outfile, std::ios::app | std::ios::binary);

    // Write configuration options first
    WRITE64(fd, numpart);
    WRITE64(fd, numhandlers);
    WRITE64(fd, numhits);
    WRITE64(fd, numclients);
    WRITE64(fd, heapsize);
    WRITE64(fd, numdocs);
    WRITE64(fd, numattr);
    WRITE64(fd, numslots);
    WRITE64(fd, distribution);
    WRITE64(fd, numqueries);
    WRITE64(fd, simtime);
    WRITE64(fd, cputime);
    WRITE64(fd, qps);
    WRITE64(fd, qps_weighted);
    WRITE64(fd, time_queue);
    WRITE64(fd, time_exec);
    WRITE64(fd, time_fetcher);
    WRITE64(fd, time_partitionqueue);
    WRITE64(fd, time_handler);
    fd.write((char *)&fraction_handler, sizeof(fraction_handler));
    for (docid_t i = 0; i < numpart; i++) {
        uint64_t val = hitstats[i];   
        WRITE64(fd, val);
        WRITE64(fd, partitionqueuetimes[i]);
        WRITE64(fd, fetchertimes[i]);
    }
    WRITE64(fd, numindices);
    for (unsigned long i = 0; i < attributeindices.size(); i++) {
        uint64_t val = attributeindices[i];
        WRITE64(fd, val);
    }
    for (unsigned long i = 0; i < partitionscheme.size(); i++) {
        uint64_t val = partitionscheme[i];
        WRITE64(fd, val);
    }
    fd.close();
}

#define READ64(fd, var) do {                \
    uint64_t tmp = 0;                       \
    (fd).read((char *)&tmp, sizeof(tmp));   \
    if ((fd).eof())                         \
        return -1;                          \
    (var) = vsimbe64toh(tmp);               \
} while (0)

int
VSimReport::importReportData(std::ifstream &fd)
{
    READ64(fd, numpart);
    READ64(fd, numhandlers);
    READ64(fd, numhits);
    READ64(fd, numclients);
    READ64(fd, heapsize);
    READ64(fd, numdocs);
    READ64(fd, numattr);
    READ64(fd, numslots);
    READ64(fd, distribution);
    READ64(fd, numqueries);
    READ64(fd, simtime);
    READ64(fd, cputime);
    READ64(fd, qps);
    READ64(fd, qps_weighted);
    READ64(fd, time_queue);
    READ64(fd, time_exec);
    READ64(fd, time_fetcher);
    READ64(fd, time_partitionqueue);
    READ64(fd, time_handler);
    fd.read((char *)&fraction_handler, sizeof(fraction_handler));
    for (docid_t i = 0; i < numpart; i++) {
        uint64_t hit;
        READ64(fd, hit);
        uint64_t pqt;
        READ64(fd, pqt);
        uint64_t ft;
        READ64(fd, ft);
        hitstats.push_back(hit);
        partitionqueuetimes.push_back(pqt);
        fetchertimes.push_back(ft);
    }
    READ64(fd, numindices);
    for (unsigned long i = 0; i < numindices; i++) {
        uint64_t val = 0;
        READ64(fd, val);
        attributeindices.push_back(val);
    }
    for (docid_t i = 0; i < numpart; i++) {
        uint64_t val;
        READ64(fd, val);
        partitionscheme.push_back(val);
    }
    return 0;
}

void
VSimReport::importReport(const char *infile)
{
    std::ifstream fd(infile, std::ios::in | std::ios::binary);
    importReportData(fd);
    fd.close();
}

std::vector<VSimReport>
importReportBatch(const char *infile)
{
    std::vector<VSimReport> reports;
    std::ifstream fd(infile, std::ios::in | std::ios::binary);
    while (!fd.eof()) {
        VSimReport report;
        if (report.importReportData(fd) != 0)
            break;
        reports.push_back(report);
    }
    fd.close();
    return (reports);
}

SlotDistribution
VSimReport::distIntToName(uint64_t val)
{
    return ((SlotDistribution)val);
}

uint64_t
VSimReport::distNameToInt(SlotDistribution sd)
{
    return ((uint64_t)sd);
}

void
VSimReport::printReport(void)
{
    std::cout << "Configuration:\t\t" << std::endl;
    std::cout << "#partitions:\t\t" << numpart << std::endl;
    std::cout << "#handlers:\t\t" << numhandlers << std::endl;
    std::cout << "#hits:\t\t\t" << numhits << std::endl;
    std::cout << "#clients:\t\t" << numclients << std::endl;
    std::cout << "heapsize:\t\t" << heapsize << std::endl;
    std::cout << "numdocs:\t\t" << numdocs << std::endl;
    std::cout << "num attributes:\t\t" << numattr << std::endl;
    std::cout << "attributes used:\t";
    for (unsigned int i = 0; i < attributeindices.size(); i++) {
        std::cout << attributeindices[i] << " ";
    }
    std::cout << std::endl;
    std::cout << "#slots/part:\t\t" << numslots << std::endl;
    std::cout << "distribution:\t\t" << slotDistEnumToText(distIntToName(distribution)) << std::endl;
    std::cout << "#queries:\t\t" << numqueries << std::endl;
    std::cout << "simtime:\t\t" << simtime << std::endl;
    std::cout << "cputime:\t\t" << cputime << std::endl;
    std::cout << std::endl;

    const unsigned int NUMCOLUMNS = 8;
    std::cout << "Results:" << std::endl;
    const char menu[NUMCOLUMNS][128] = { "Queries/sec", "Weighted_Q/s", "time_queue", "time_execution", "time_fetcher", "time_partitionqueue", "time_handler", "\%handler" };
    std::vector<int> widths(NUMCOLUMNS);
    for (unsigned int i = 0; i < NUMCOLUMNS; i++) {
        widths[i] = strlen(menu[i]);
        if (i > 0)
            widths[i]++;
        std::cout << std::setw(widths[i]) << menu[i];
    }
    std::cout << std::endl;
    std::cout << std::setw(widths[0]) << qps;
    std::cout << std::setw(widths[1]) << qps_weighted;
    std::cout << std::setw(widths[2]) << time_queue;
    std::cout << std::setw(widths[3]) << time_exec;
    std::cout << std::setw(widths[4]) << time_fetcher;
    std::cout << std::setw(widths[5]) << time_partitionqueue;
    std::cout << std::setw(widths[6]) << time_handler;
    std::cout << std::setw(widths[7]) << std::setprecision(2) << fraction_handler;
    std::cout << std::endl;
    std::cout << std::endl;
    std::cout << "Per partition stats: " << std::endl;
    for (docid_t i = 0; i < numpart; i++) {
        std::cout << i << " H: " << hitstats[i] << " P: " << partitionqueuetimes[i] << " F: " << fetchertimes[i] << std::endl;
    }
}

} // end vsim
