#include <iostream>
#include <iomanip>
#include <inttypes.h>
#include <assert.h>
#include <numeric>
#include <cstring>

#include "types.h"
#include "atomic.h"
#include "vsimstat.h"
#include "vsimconfig.h"
#include "misc.h"

namespace vsim {

using namespace vespalib;

VSimStat::VSimStat(VSimConfig &vcfg) :
    _vcfg(vcfg),
    _hs(vcfg.getNumPartitions(), 0),
    _parths(vcfg.getNumPartitions(), 0),
    _numfinished(0),
    _fetchertimes(vcfg.getNumPartitions(), 0),
    _partitionqueuetimes(vcfg.getNumPartitions(), 0),
    _totalmaxfetchertime(0),
    _totalqueryqueuetime(0),
    _totalhandlertime(0)
{ }

void
VSimStat::report(std::vector<uint64_t> &fetchertimes, std::vector<uint64_t> &partitionqueuetimes, uint64_t maxfetchertime, uint64_t queuetime, uint64_t handlertime)
{
    Atomic::postInc(&_numfinished);
    for (partid_t i = 0; i < fetchertimes.size(); i++) {
        Atomic::add(&_fetchertimes[i], fetchertimes[i]);
        Atomic::add(&_partitionqueuetimes[i], partitionqueuetimes[i]);
    }
    Atomic::add(&_totalmaxfetchertime, maxfetchertime);
    Atomic::add(&_totalqueryqueuetime, queuetime);
    Atomic::add(&_totalhandlertime, handlertime);
}

void
VSimStat::reportHits(HitStat hs)
{
    if (hs.size() != _hs.size()) {
        std::cout << "Cannot merge stats of different size, ours: " << _hs.size() << " real: " << hs.size() << std::endl;
        return;
    }
    for (unsigned int i = 0; i < _hs.size(); i++) {
        _hs[i] += hs[i];
    }
    _hslock.lock();
    for (unsigned int i = 0; i < _parths.size(); i++) {
        _parths[i] += hs[i];
    }
    _hslock.unlock();
}

#include <cstdlib>

/**
 * Update partition spec based on hit statistics.
 */
PartitionSpecSP
VSimStat::updatePartitionSpec(const PartitionSpecSP &ps)
{
    _hslock.lock();
    HitStat hs(_parths);
    _hslock.unlock();

    assert((*ps).size() == hs.size());

    docid_t totalhit = 0;
    for (unsigned int i = 0; i < hs.size(); i++) {
        hs[i]++;
        totalhit += hs[i];
    }

    docid_t numslots = _vcfg.getNumSlots();
    docid_t idealhit = totalhit / hs.size();
    docid_t slothit = idealhit / numslots;
    std::vector<docid_t> numslots_per(hs.size(), numslots);
    PartitionSpecSP newspec(new PartitionSpec(*ps));

    for (unsigned int i = 0; i < hs.size(); i++) {
        // We need more slots, take one from our neighbour
        if (hs[i] < idealhit) {
            docid_t remainder = idealhit - hs[i];
            docid_t remainder_slots = remainder / slothit;
            for (unsigned int j = hs.size() - 1; j > i && remainder_slots > 0; j--) {
                PartitionRange &mine((*newspec)[j]);
                docid_t available_slots = MIN(mine.size(), remainder_slots);
                // Move these backwards
                PartitionRange &prev((*newspec)[j - 1]);
                prev.insert(prev.end(), mine.begin(), (mine.begin() + available_slots));
                mine.erase(mine.begin(), (mine.begin() + available_slots));
                numslots_per[j] -= available_slots;
                numslots_per[j - 1] += available_slots;
            }
       
        } else if (hs[i] > idealhit) {
            docid_t remainder = hs[i] - idealhit;
            docid_t remainder_slots = remainder / slothit;
            for (unsigned int j = i; j < hs.size() - 1; j++) {
               PartitionRange &mine((*newspec)[j]);
               PartitionRange &next((*newspec)[j + 1]);
               docid_t available_slots = MIN(remainder_slots, mine.size());
               // Move these forward
               next.insert(next.begin(), (mine.end() - available_slots), mine.end());
               mine.erase((mine.end() - available_slots), mine.end());

               numslots_per[j] -= available_slots;
               numslots_per[j + 1] += available_slots;
            }
        }
        hs[i] = idealhit;
    }
    _hslock.lock();
    _parths = hs;
    _hslock.unlock();
    return (newspec);
}

void
VSimStat::sleepUntilFinished(void)
{
    VSimTime now;
    now.recordTime();
    unsigned long diff = (now.getTimeSeconds() - _startSim.getTimeSeconds());
    std::cout << "Sleeping for " << (_vcfg.getSimTime() - diff) << " seconds\n";
    sleep(_vcfg.getSimTime() - diff);
}

void
VSimStat::startSimulation(void)
{
    _startSim.recordTime();
    getrusage(RUSAGE_SELF, &_rusage_start);
}

void
VSimStat::stopSimulation(void)
{
    _stopSim.recordTime();
    getrusage(RUSAGE_SELF, &_rusage_end);
}

VSimReport
VSimStat::generateReport(const PartitionSpec &spec)
{
    uint64_t totalfetchertime = 0;
    uint64_t totalpartqueuetime = 0;
    for (unsigned long i = 0; i < _fetchertimes.size(); i++) {
        totalpartqueuetime += _partitionqueuetimes[i];
        totalfetchertime += _fetchertimes[i];
    }
    totalpartqueuetime /= _partitionqueuetimes.size();
    totalfetchertime /= _fetchertimes.size();

    VSimTime usertime_start(_rusage_start.ru_utime);
    VSimTime usertime_end(_rusage_end.ru_utime);
    VSimTime systime_start(_rusage_start.ru_stime);
    VSimTime systime_end(_rusage_end.ru_stime);

    VSimReport report;
    report.numpart = _fetchertimes.size();
    report.numhandlers = _vcfg.getMaxConnections();
    report.numhits = _vcfg.getNumHits();
    report.numclients = _vcfg.getNumClients();
    report.heapsize = _vcfg.getHitHeapSize();
    report.numdocs = _vcfg.getMaxDocId();
    report.numattr = _vcfg.getNumAttributes();
    report.numslots = _vcfg.getNumSlots();
    report.distribution = VSimReport::distNameToInt(_vcfg.getSlotDistribution());
    report.numqueries = _numfinished;
    report.simtime = _startSim.getDiffMilliSeconds(_stopSim) / 1000;
    report.cputime = (usertime_start.getDiffMilliSeconds(usertime_end) + systime_start.getDiffMilliSeconds(systime_end)) / 1000;
    report.qps = (_numfinished / report.simtime);
    report.qps_weighted = (_numfinished / report.cputime);
    report.time_queue = (_totalqueryqueuetime / _numfinished);
    report.time_exec = (_totalhandlertime / _numfinished);
    report.time_fetcher = (_totalmaxfetchertime / _numfinished);
    report.time_partitionqueue = (totalpartqueuetime / _numfinished);
    report.time_handler = report.time_exec - report.time_fetcher - report.time_partitionqueue;
    report.fraction_handler = (double)100.0 * (double)report.time_handler / (double)report.time_exec;
    report.hitstats = _hs;
    for (docid_t i = 0; i < report.hitstats.size(); i++) {
        report.partitionqueuetimes.push_back(_partitionqueuetimes[i] / _numfinished);
        report.fetchertimes.push_back(_fetchertimes[i] / _numfinished);
    }
    report.attributeindices = _vcfg.getAttributeIndices();
    report.numindices = report.attributeindices.size();
    for (docid_t i = 0; i < spec.size(); i++) {
        const PartitionRange &pr(spec[i]);
        report.partitionscheme.push_back(pr.size());
    }
    return (report);
}

} // end vsim
