#include <iostream>
#include <algorithm>
#include <sys/mman.h>
#include <boost/random/linear_congruential.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/variate_generator.hpp>
#include "vsimconfig.h"
#include "vsimstore.h"

#define CACHE_LINE_SIZE 64

namespace vsim {

typedef boost::minstd_rand base_generator_type;

VSimStore::VSimStore(const VSimConfig &vcfg) :
    _maxdocid(vcfg.getMaxDocId()),
    _attrindex(vcfg.getAttributeIndices())
{
    // Create memory area for all attributes
    for (unsigned int i = 0; i < vcfg.getNumAttributes(); i++) {
        attrptr_t astore = 0;
        if (posix_memalign((void **)&astore, CACHE_LINE_SIZE, (_maxdocid * sizeof(attr_t))) != 0) {
            std::cout << "Error allocating aligned memory for attributes" << std::endl;
            exit(-1);
        }
        _attributes.push_back(astore);
    }
}

void
VSimStore::memlock(void)
{
    for (unsigned int i = 0; i < _attributes.size(); i++) {
        attrptr_t astore = _attributes[i];
        mlock(astore, (_maxdocid * sizeof(attr_t)));
    }
}

void
VSimStore::memunlock(void)
{
    for (unsigned int i = 0; i < _attributes.size(); i++) {
        attrptr_t astore = _attributes[i];
        munlock(astore, (_maxdocid * sizeof(attr_t)));
    }
}

void
VSimStore::seed(unsigned long s)
{
    seedRange(s, 0, _maxdocid);
}

void
VSimStore::seedRange(unsigned long s, docid_t start, docid_t end)
{
    base_generator_type generator(s);
    boost::uniform_int <attr_t> dist(1, MAXRANK);
    boost::variate_generator<base_generator_type&, boost::uniform_int<attr_t> > vgen(generator, dist);

    for (docid_t i = start; i < end; i++) {
        for (unsigned int j = 0; j < _attributes.size(); j++) {
            attrptr_t astore = _attributes[j];
            astore[i] = vgen();
        }
    }
}

VSimStore::~VSimStore()
{
    // Create memory area for all attributes
    for (unsigned int i = 0; i < _attributes.size(); i++) {
        attr_t *astore = _attributes[i];
        free(astore);
    }
}

rank_t
VSimStore::getRank(docid_t docid) const
{
    const unsigned int numrounds = 20;
    rank_t rank = 0;

    // Busy loop tuned to blow some cycles here
    for (unsigned int i = 0; i < numrounds; i++) {
        for (unsigned long a = 0; a < _attrindex.size(); a++) {
            attrptr_t astore = _attributes[_attrindex[a]];
            rank += astore[docid] + i;
        }
    }
    return (rank);

}

} // end vsim
