#!/usr/bin/env python
import struct
import stat
import os
import sys
import math

# The simreport module is able to read simulator output reports generated by
# Vsim and provides an object oriented interface towards the reports. Moreover,
# it contains support for generating mean values of multiple runs together with
# standard deviations of these values. The script may also be run standalone
# with a filename to the sample file as argument.
def ReadInt(handle):
    return struct.unpack_from('>Q', handle.read(8))[0]

def ReadDouble(handle):
    return struct.unpack_from('d', handle.read(8))[0]

class SimReport:
    def __init__(self, reportNum, filename=None):
        self.numpart = 0
        self.numhandlers = 0
        self.numhits = 0
        self.numclients = 0
        self.heapsize = 0
        self.numdocs = 0
        self.numattr = 0
        self.numslots = 0
        self.distribution = 0
        self.numqueries = 0
        self.simtime = 0
        self.cputime = 0
        self.qps = [0]
        self.qps_mean = 0
        self.qps_stddev = 0
        self.qps_weighted = 0
        self.time_queue = 0
        self.time_exec = 0
        self.time_fetcher = 0
        self.time_partitionqueue = 0
        self.time_handler = 0
        self.fraction_handler = 0.0
        self.hitstat = []
        self.partitionqueuetime = []
        self.fetchertimes = []
        self.numindices = 0
        self.attributeindices = []
        self.partitionscheme = []

        self.reportnum = reportNum
        if filename == None:
            return
        handle = open(filename, mode='rb')
        self.importData(handle)
        handle.close()

    def importData(self, handle):
        self.numpart = ReadInt(handle)
        self.numhandlers = ReadInt(handle)
        self.numhits = ReadInt(handle)
        self.numclients = ReadInt(handle)
        self.heapsize = ReadInt(handle)
        self.numdocs = ReadInt(handle)
        self.numattr = ReadInt(handle)
        self.numslots = ReadInt(handle)
        self.distribution = ReadInt(handle)
        self.numqueries = ReadInt(handle)
        self.simtime = ReadInt(handle)
        self.cputime = ReadInt(handle)
        self.qps[0] = ReadInt(handle)
        self.qps_weighted = ReadInt(handle)
        self.time_queue = ReadInt(handle)
        self.time_exec = ReadInt(handle)
        self.time_fetcher = ReadInt(handle)
        self.time_partitionqueue = ReadInt(handle)
        self.time_handler = ReadInt(handle)
        self.fraction_handler = ReadDouble(handle)
        self.hitstat = []
        self.partitionqueuetime = []
        self.fetchertimes = []
        self.attributeindices = []
        self.partitionscheme = []
        for i in range(self.numpart):
            self.hitstat.append(ReadInt(handle))
            self.partitionqueuetime.append(ReadInt(handle))
            self.fetchertimes.append(ReadInt(handle))
        self.numindices = ReadInt(handle)
        for i in range(self.numindices):
            self.attributeindices.append(ReadInt(handle))
        for i in range(self.numpart):
            self.partitionscheme.append(ReadInt(handle))

    def equalConfig(self, report):
        return (self.numpart == report.numpart and self.numhandlers == report.numhandlers and self.numhits == report.numhits and self.numclients == report.numclients and self.heapsize == report.heapsize and self.numdocs == report.numdocs and self.numattr == report.numattr and self.numslots == report.numslots and self.distribution == report.distribution and self.numindices == report.numindices) 

    def add(self, report):
        self.qps.extend(report.qps)
        self.qps_weighted += report.qps_weighted
        self.time_queue += report.time_queue
        self.time_exec += report.time_exec
        self.time_fetcher += report.time_fetcher
        self.time_partitionqueue += report.time_partitionqueue
        self.time_handler += report.time_handler
        self.fraction_handler += report.fraction_handler
        for i in range(min(self.numpart, report.numpart)):
            self.hitstat[i] += report.hitstat[i]
            self.partitionqueuetime[i] += report.partitionqueuetime[i]
            self.fetchertimes[i] += report.fetchertimes[i]
            self.partitionscheme[i] += report.partitionscheme[i]

    def printreport(self):
        print self.numhandlers, self.numpart, self.numhits, self.qps_mean, self.qps_stddev, self.time_queue, self.time_exec, self.time_fetcher, self.time_partitionqueue

    def average(self, howmany):
        self.qps_mean = sum(self.qps) / howmany
        for num in self.qps:
            self.qps_stddev += math.pow(num - self.qps_mean, 2)
        self.qps_stddev = math.sqrt(self.qps_stddev)
        self.qps_weighted /= howmany
        self.time_queue /= howmany
        self.time_exec /= howmany
        self.time_fetcher /= howmany
        self.time_partitionqueue /= howmany
        self.time_handler /= howmany
        self.fraction_handler /= howmany
        for i in range(self.numpart):
            self.hitstat[i] /= howmany
            self.partitionqueuetime[i] /= howmany
            self.fetchertimes[i] /= howmany
            self.partitionscheme[i] /= howmany

def importReportBatch(filename):
    s = os.stat(filename)
    size = s[stat.ST_SIZE]
    handle = open(filename, 'rb')
    reports = []
    i = 0
    while handle.tell() < size:
        report = SimReport(i)
        report.importData(handle)
        reports.append(report)
        i += 1
    handle.close()
    return reports

def averageReports(reportlist):
    avglist = []
    numavg = []
    for report in reportlist:
        found = False
        for i in range(len(avglist)):
            if report.equalConfig(avglist[i]):
                avglist[i].add(report)
                numavg[i] += 1
                found = True
                break
        if not found:
            avglist.append(report)
            numavg.append(1)
    for i in range(len(avglist)):
        avglist[i].average(numavg[i])
    return avglist

def importAvgReports(filename):
    return (averageReports(importReportBatch(filename)))

if (len(sys.argv) > 1):
    reports = importReportBatch(sys.argv[1])
    for report in reports:
        report.printreport()
